import ast
import collections
import importlib.resources
import os.path
import urllib.parse

import batou.lib.git
import batou.lib.mercurial
from batou.component import Component
from batou.lib.file import Content, File


class Source(Component):
    dist_sources = "[]"
    sources = "[]"

    hg_hostfingerprints = {}
    additional_hgrc_content = ""

    vcs_update = "True"

    def _eval_attr(self, *names):
        for name in names:
            if isinstance(getattr(self, name), str):
                setattr(self, name, ast.literal_eval(getattr(self, name)))

    def configure(self):
        self._eval_attr(
            "dist_sources", "sources", "vcs_update", "hg_hostfingerprints"
        )

        self.hg_hostfingerprints = collections.OrderedDict(
            x for x in sorted(self.hg_hostfingerprints.items())
        )

        self.provide("source", self)

        additional_hgrc = os.path.join(self.defdir, "hgrc")
        if not self.additional_hgrc_content and os.path.exists(additional_hgrc):
            self |= Content("additional_hgrc", source=additional_hgrc)
            self.additional_hgrc_content = self._.content

        self.hgrc = File(
            "~/.hgrc",
            source=str(
                importlib.resources.files("batou_scm").joinpath(
                    "resources/hgrc"
                )
            ),
        )
        self += self.hgrc

        self.distributions = dict(
            self.add_clone(url) for url in self.dist_sources
        )
        self.clones = self.distributions.copy()
        self.clones.update(self.add_clone(url) for url in self.sources)

    VCS = {
        "hg": {
            "component": batou.lib.mercurial.Clone,
            "default-branch": "default",
        },
        "git": {
            "component": batou.lib.git.Clone,
            "default-branch": "main",
        },
    }

    def add_clone(self, url):
        parts = urllib.parse.urlsplit(url)
        vcs, _, scheme = parts.scheme.partition("+")
        if not vcs:
            raise ValueError(
                "Malformed VCS URL %r, need <vcs>+<protocol>://<url>, "
                "e.g. hg+ssh://hg@bitbucket.org/gocept/batou" % url
            )
        url = urllib.parse.urlunsplit((scheme,) + parts[1:])

        url, _, parameter_list = url.partition(" ")
        parameters = {}
        parameters["target"] = list(filter(bool, url.split("/")))[-1]
        parameters.update(x.split("=") for x in parameter_list.split())
        if "branch" not in parameters and "revision" not in parameters:
            parameters["branch"] = self.VCS[vcs]["default-branch"]

        clone = self.VCS[vcs]["component"]
        self += clone(url, vcs_update=self.vcs_update, **parameters)
        return parameters["target"], self._



import os.path

import pytest
from batou.lib.mercurial import Clone

from batou_scm.source import Source


@pytest.fixture
def source(root):
    source = Source(
        dist_sources=repr(
            [
                "hg+https://example.com/foo",
                "hg+https://example.com/bar branch=BAR",
            ]
        ),
        sources=repr(["hg+https://example.com/baz revision=BAZ"]),
    )
    root.component += source
    root.component.configure()
    return source


def test_all_clones_are_configured_as_subcomponents(source):
    clones = [s for s in source.sub_components if isinstance(s, Clone)]
    assert set(clones) == set(source.clones.values())
    assert set(clone.url for clone in clones) == set(
        [
            "https://example.com/foo",
            "https://example.com/bar",
            "https://example.com/baz",
        ]
    )


def test_all_targets_are_derived_from_clone_urls(source):
    assert {
        clone.url: clone.target for clone in list(source.clones.values())
    } == {
        "https://example.com/foo": source.map("foo"),
        "https://example.com/bar": source.map("bar"),
        "https://example.com/baz": source.map("baz"),
    }


def test_distributions_are_collected_separately(source):
    assert {
        clone.url: clone.target for clone in list(source.distributions.values())
    } == {
        "https://example.com/foo": source.map("foo"),
        "https://example.com/bar": source.map("bar"),
    }


def test_branches_are_read_from_parameters(source):
    assert source.clones["bar"].branch == "BAR"


def test_branch_is_set_to_default_if_not_given(source):
    assert source.clones["foo"].branch == "default"


def test_branch_is_not_set_if_revision_given(source):
    assert source.clones["baz"].branch is None


def test_revisions_are_read_from_parameters(source):
    assert source.clones["foo"].revision is None
    assert source.clones["bar"].revision is None
    assert source.clones["baz"].revision == "BAZ"


def test_hg_hostfingerprints_are_sorted(source):
    source.hg_hostfingerprints["code.gocept.com"] = "fpr3"
    source.hg_hostfingerprints["bitbucket.org"] = "fpr1"
    source.hg_hostfingerprints["example.com"] = "fpr2"
    source.configure()
    assert source.hgrc.content.startswith(b"""\
[hostfingerprints]
bitbucket.org = fpr1
code.gocept.com = fpr3
example.com = fpr2
""")


def test_additional_hgrc_content_is_taken_from_file_if_present(source):
    open(os.path.join(source.defdir, "hgrc"), "w").write("foo")
    source.configure()
    assert b"foo" in source.hgrc.content


def test_malformed_vcs_url_raises_key_error(root):
    source = Source(
        dist_sources=repr(["https://example.com/foo"]),
    )
    with pytest.raises(KeyError):
        root.component += source


def test_vcs_url_without_vcs_prefix_raises_value_error(root):
    source = Source(
        dist_sources=repr(["+example.com/foo"]),
    )
    with pytest.raises(ValueError):
        root.component += source



import configparser

import mock
import pytest
from batou import UpdateNeeded
from batou.lib.file import Directory, File

from batou_scm.buildout import Buildout, BuildoutWithVersionPins
from batou_scm.source import Source


@pytest.fixture
def buildout(root):
    source = Source(
        dist_sources=repr(
            [
                "hg+https://example.com/foo",
                "hg+https://example.com/bar",
            ]
        )
    )
    root.component += source
    root.component += Buildout(config=File("buildout.cfg", content=""))
    root.component.configure()
    return root.component._


@pytest.fixture(scope="function")
def buildout_with_version_pins(root):
    source = Source(
        dist_sources=repr(
            [
                "hg+https://example.com/foo",
                "hg+https://example.com/bar",
            ]
        ),
        sources=repr(["hg+https://example.com/ver target=versions"]),
    )
    root.component += source
    source.versions = source.clones["versions"]
    root.component += BuildoutWithVersionPins(
        config=File("buildout.cfg", content="")
    )
    root.component.configure()
    return root.component._


def read_config(content):
    config = configparser.ConfigParser()
    config.read_string(content.decode("ascii"))
    return config


def test_buildout_doesnt_need_source_to_be_configured(root):
    root.component += Buildout(config=File("buildout.cfg", content=""))
    root.component.configure()
    assert b".batou-shared-eggs" in root.component._.overrides.content
    with pytest.raises(UpdateNeeded):
        root.component._.verify()


def test_dist_paths_are_listed_as_develop_paths_in_overrides(buildout):
    config = read_config(buildout.overrides.content)
    assert set(
        [
            buildout.source.map("bar"),
            buildout.source.map("foo"),
        ]
    ).issubset(config.get("buildout", "develop +").split())


def test_no_dist_sources_configured_does_not_break(root):
    source = Source()
    root.component += source
    root.component += Buildout(config=File("buildout.cfg", content=""))
    root.component.configure()
    buildout = root.component._
    assert b"develop +=" not in buildout.overrides.content
    assert buildout.source is not None


def test_develop_paths_in_overrides_have_defined_order(buildout):
    config = read_config(buildout.overrides.content)
    develop_paths = config.get("buildout", "develop +").split()
    assert develop_paths == sorted(develop_paths)


def test_dist_versions_are_unpinned_in_overrides(buildout):
    config = read_config(buildout.overrides.content)
    assert config.get("versions", "foo") == ""
    assert config.get("versions", "bar") == ""


def test_versions_of_setuptools_and_buildout_are_pinned_explicitly(buildout):
    config = read_config(buildout.overrides.content)
    assert config.get("versions", "setuptools") == str(buildout.setuptools)
    assert config.get("versions", "zc.buildout") == str(buildout.version)


def test_version_pins_in_overrides_have_defined_order(buildout):
    config = read_config(buildout.overrides.content)
    pins = config.items("versions")
    assert pins == sorted(pins)


def test_shared_egg_dir_is_created_for_service_user(buildout):
    config = read_config(buildout.overrides.content)
    eggs_dir = config.get("buildout", "eggs-directory")
    assert eggs_dir == buildout.map("~/" + eggs_dir.split("/")[-1])
    assert any(
        component.path == eggs_dir
        for component in buildout.sub_components
        if isinstance(component, Directory)
    )


def test_verify_checks_source(buildout):
    with mock.patch("batou.lib.buildout.Buildout.verify"):
        with mock.patch("batou.lib.mercurial.Clone.has_changes") as has_changes:
            has_changes.return_value = False
            buildout.source.assert_no_changes = mock.Mock()
            buildout.verify()
            assert buildout.source.assert_no_changes.called
            assert has_changes.called


def test_verify_result_is_cached(buildout):
    with mock.patch("batou.lib.buildout.Buildout.verify"):
        with mock.patch("batou.lib.mercurial.Clone.has_changes"):
            with mock.patch.object(
                buildout.source, "assert_no_subcomponent_changes"
            ) as ansc:
                ansc.side_effect = UpdateNeeded
                with pytest.raises(UpdateNeeded):
                    buildout.verify()
                with pytest.raises(UpdateNeeded):
                    buildout.verify()
                assert 1 == ansc.call_count


def test_buildout__BuildoutWithVersionPins__verify__1(
    buildout_with_version_pins,
):
    """`verify()` requests update on changed versions."""
    buildout = buildout_with_version_pins
    buildout.__update_needed = None
    with (
        mock.patch("batou.lib.buildout.Buildout.verify"),
        mock.patch.object(
            buildout.source.clones["versions"], "has_changes", return_value=True
        ),
        mock.patch("batou.lib.mercurial.Clone.has_changes", return_value=False),
        mock.patch(
            "batou.lib.mercurial.Clone.has_outgoing_changesets",
            return_value=False,
        ),
    ):
        with pytest.raises(UpdateNeeded):
            buildout.verify()


def test_buildout__BuildoutWithVersionPins__verify__2(
    buildout_with_version_pins,
):
    """`verify()` requests update on outgoing changes in versions."""
    buildout = buildout_with_version_pins
    buildout.__update_needed = None
    with (
        mock.patch("batou.lib.buildout.Buildout.verify"),
        mock.patch.object(
            buildout.source.clones["versions"],
            "has_changes",
            return_value=False,
        ),
        mock.patch(
            "batou.lib.mercurial.Clone.has_outgoing_changesets",
            return_value=True,
        ),
        mock.patch("batou.lib.mercurial.Clone.has_changes", return_value=False),
    ):
        with pytest.raises(UpdateNeeded):
            buildout.verify()
